//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// iSCSI Login Reference Model
//
// Matthew Burbridge
// Senior Development Engineer
// Hewlett Packard Ltd.
// matthewb@bri.hp.com
//
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// This programme models the iSCSI Login Phase proposal of 14-Aug-2001
// It models the behaviour of both the initiator and the target when
// confronted by various user entered parameters.
//
// Notes:
//
// Please note that there is NO parameter value processing except for
// SecurityContextComplete=yes (or SCC=yes) and SessionType=... All
// other parameter values can be anything and are not checked..
//
// This is NOT the world's (or mine for that matter) efficient code as
// the main object is to validate the iSCSI login phase.
//
// Built off of iSCSI draft 0.7 spec.
//
// OpParmReset is NOT supported (no longer required)
//
// To reduce complexity of the model implementation the following parameters
// are NOT supported.
//
// Full Security negotiation
// SendTargets + Response
// AccessID
// Glen-Turner Vendor Specific Keys
//
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// Compliation Instructions:
//
// Linux: gcc -Wall -o LoginReference LoginReference.cpp
// Windows: Use the MS Visual C++ Development Environment of similar
// Others: Its a simple C file (with C++ comments!) so should build on anything.
//
// Known problems:
//
// Session type must be in the Login command (proposal states it is optional)
//
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// SSC MUST not be in the initial login command!

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// OPTIONS++
//
// Choose which one you want to use!
#define SCC		"SCC="
//#define SCC		"SecurityContextComplete="
//
#define TARGET_WANTS_SECUIRTY	0	// 0 = no or 1 = yes!
//
#define ECHOINPUT 1			// 0 = no or 1 = yes For input redirect

// OPTIONS--
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// -----------
// Definitions
// -----------
#define MAX_PARAM_LENGTH ( 63 + 1 + 255 + 1 ) // Used (as I'm lazy) for all text input!
#define MAX_PARAMETERS	100	// Should be more than enough

typedef enum { LOGIN, TEXT, REJECT } PDUTYPE;
typedef enum { TWaitingForLogin,               TSecurity, TOperational, TLoggedIn } TSTATE;
typedef enum { INotLoggedIn,     IPreSecurity, ISecurity, IOperational, ILoggedIn } ISTATE;
typedef enum { Normal, Discovery, CopyManager, Boot } SESSIONTYPE;
typedef char PARAMS[MAX_PARAMETERS][MAX_PARAM_LENGTH];

typedef struct
{
	PDUTYPE Type;
	int		FBit;
	PARAMS	Parameters;
	int		LoginStatus;  

} PDU_t;

typedef enum { FREE, OUT_REQUEST, IN_REQUEST, DONE, STATUS } PSTATE;

typedef struct
{
	PSTATE	State;
	char	Text[MAX_PARAM_LENGTH];

} PARAMTYPE;

// +++++++++++++++++++
// Function Prototypes
// +++++++++++++++++++
static void ReadString( const char *text, char *str );
static int ReadBoolean( const char *text );
static void SendCommand( PDU_t *command, PDU_t *response );
static void TargetProcessCommand( PDU_t *command, PDU_t *response );
static int IsSecurityParameter( char *parameter );
static int ParametersContainSecurity( PARAMS parameters );
static int IsOperationalParameter( char *parameter );
static int ParametersContainOperational( PARAMS parameters );
static void PrintResponse( PDU_t *response );
static void ReadParameters( PARAMS Parameters, PARAMTYPE *ParamSet );
static int GetParameter( PARAMS parameters, char *parameter, char *value );
static SESSIONTYPE GetSessionType( char *value );
static void Initiator( void );
static void AddOutParameter( char *Parameter, PARAMTYPE *ParamSet );
static void AddInParameter( char *Parameter, PARAMTYPE *ParamSet );
static void ProcessInParameters( PARAMS Parameters, PARAMTYPE *ParamSet );
static int CheckAllRequestsSatisfied( PARAMTYPE *ParamSet );
static int AreAllRequestsSatisfied( PARAMTYPE *ParamSet );
static int AreAllRepliesSatisfied( PARAMTYPE *ParamSet );
static void ChangeParameterState( char *Parameter, PARAMTYPE *ParamSet, PSTATE );
static void DoExit( int code );
static int GetNoOfParameters( PARAMS parameters );
static int SetFBit( int *FBit );
static void SendOperationalTextCommand( PDU_t *command, PDU_t *response );
static void SendSecurityTextCommand( PDU_t *command, PDU_t *response );
static void LoginResponse( PDU_t *response, int Status );


static ISTATE IState = INotLoggedIn;		// State of the target state machine (with initial state)
static TSTATE TState = TWaitingForLogin;	// State of the target state machine (with initial state)

static PARAMTYPE	TargetParamSet[MAX_PARAMETERS];		// Set of parameters used by the target to track them
static PARAMTYPE	InitiatorParamSet[MAX_PARAMETERS];	// Set of parameters used by the target to track them

static int			TargetFBit = 0;			// FBit owned by the target
static int			InitiatorFBit = 0;		// FBit owned by the initiator
static int			I_InitiatorSCC = 0;		// Initiator SCC status as seen by the initiator
static int			I_TargetSCC = 0;		// Target SCC status as seen by the initiator
static int			T_InitiatorSCC = 0;		// Initiator SCC status as seen by the target
static int			T_TargetSCC = 0;		// Target SCC status as seen by the target

static int TargetWantsSecurity = TARGET_WANTS_SECUIRTY;
	
const char *SecurityParameters[] =
{
	"DataDigest=",
	"HeaderDigest=",
	"AuthMethod=",
	NULL
};

const char *OperationalParameters[] =
{
	"MaxConnections=",
	"FMarker=",
	"RFMarkInt=",
	"SFMarkInt=",
	"InitialR2T=",
	"BiDiInitialR2T=",
	"ImmediateData=",
	"DataPDULength=",
	"FirstBurstSize=",
	"LogoutLoginMinTime=",
	"LogoutLoginMaxTime=",
	"MaxOutstandingR2T=",
	"DataOrder=",
	"DataDeliveryOrder=",
	"CommandReplaySupport=",
	"CommandFailoverSupport=",
	NULL
};

const char *PDUNames[] = { "Login", "Text", "Reject" };

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Main!
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
int main ( )
{
	printf( "Login Reference Model for iSCSI\n" );
	printf( "Matthew Burbridge (HP)\n" );
	printf( "Hewlett Packard 2001\n" );

	Initiator();

	return 0;
}

// +++++++++++++++++++++++++++
// The Initiator State MAchine
// +++++++++++++++++++++++++++
static void Initiator( void )
{
	PDU_t		command;
	PDU_t		response;

	// ------------------------
	// Send the initial command
	// ------------------------
	command.Type = LOGIN;
	printf( "\nCOMMAND:\n" );
	ReadParameters( command.Parameters, InitiatorParamSet );
	command.FBit = SetFBit( &InitiatorFBit );
	SendCommand( &command, &response );

	// These parameters are not negotiated so set them to done
	ChangeParameterState( "InitiatorName=", InitiatorParamSet, DONE );
	ChangeParameterState( "SessionType=", InitiatorParamSet, DONE );
	ChangeParameterState( "TargetName=", InitiatorParamSet, DONE );

	IState = IPreSecurity;

	for ( ;; )
	{
		switch ( IState )
		{
		case INotLoggedIn:
			printf( "ERROR : Should never reach the INotLoggedInState\n" );
			DoExit( 1 );
			break;

		case IPreSecurity:
			if ( response.Type == LOGIN )
			{
				// Handle a login response
				ProcessInParameters( response.Parameters, InitiatorParamSet );
				if ( response.LoginStatus == 0 )
				{
					if ( command.FBit == 0 )
					{
						printf( "ERROR: Target sent login response before initiator had set FBit to 1\n" );
						DoExit( 1 );
					}
					IState = ILoggedIn;
				}
				else
				{
					printf( "Login Failed %04X\n", response.LoginStatus );
					DoExit( 1 );
				}
			}
			else if ( ParametersContainSecurity( response.Parameters ) )
			{
				// Target wants to enter security phase
				IState = ISecurity;
			}
			else
			{
				// Target does not want to enter operational phase (and neither does initiator
				IState = IOperational;
			}
			break;			

		case ISecurity:
			if ( ParametersContainOperational( response.Parameters ) )
			{
				printf( "Error: MUST NOT have Operational Parameters in security phase\n" );
				DoExit( 1 );
			}

			ProcessInParameters( response.Parameters, InitiatorParamSet );
			if ( GetParameter( response.Parameters, SCC, NULL ) == 1 )
			{
				// Target has reported SCC: Mark its passing.
				ChangeParameterState( SCC, InitiatorParamSet, STATUS );
				I_TargetSCC = 1;
			}
			else if ( I_TargetSCC == 1 )
			{
				printf( "ERROR: Target MUST report SecurityContextComplete=yes when Security Phase Complete\n" );
				DoExit ( 1 );
			}

			if ( ( I_InitiatorSCC == 1 ) &&
				 ( I_TargetSCC == 1 ) &&
				 ( GetNoOfParameters( response.Parameters ) == 1 )  // Must be the only parameter in the response
			   )
			{
				// All criteria met for exiting security phase
				printf ( "INITIATOR: Security Context Complete\n" );

				if ( ( command.FBit == 1 ) && ( response.FBit == 1 ) )
				{
					// No more negotiation to be done: So logged in!
					// Should I have extra checking here?
					IState = ILoggedIn;
				}
				else
				{
					// Need to go to operational phase: So send the first command
					IState = IOperational;
					SendOperationalTextCommand( &command, &response );
				}
			}
			else
			{
				// Still in security phase: carry on...
				SendSecurityTextCommand( &command, &response );
			}
			break;
			
		case IOperational:
			if ( ParametersContainSecurity( response.Parameters ) )
			{
				printf( "ERROR: MUST NOT have Security Parameters in operational phase\n" );
				DoExit( 1 );
			}
			
			ProcessInParameters( response.Parameters, InitiatorParamSet );

			if ( response.Type == LOGIN )
			{
				// We've logged in
				// Do I need extra checking here?
				IState = ILoggedIn;
			}
			else
			{
				// Still in operational phase so carry on...
				SendOperationalTextCommand( &command, &response );
			}
			break;
			
		case ILoggedIn:
			printf( "Success\n" );
			DoExit( 0 );
			break;

		default:
			// Just in case!
			printf( "ERROR : Invalid IState\n" );
			DoExit ( 1 );
			break;
		}	
	}
}

// +++++++++++++++++++++++++++
// The Target State Machine
// +++++++++++++++++++++++++++
void TargetProcessCommand( PDU_t *command, PDU_t *response )
{
	char			InitiatorName[MAX_PARAM_LENGTH];
	char			TargetName[MAX_PARAM_LENGTH];
	char			Value[MAX_PARAM_LENGTH];
	SESSIONTYPE		SessionType;

	for (;;)
	{
		switch ( TState )
		{
		case TWaitingForLogin:
			if ( command->Type != LOGIN )
			{
				printf( "Kill Connection: Command is not a Login\n" );
				DoExit ( 0 );
			}

			// --------------------------
			// Start Initial LOGIN checks
			// --------------------------
			if ( ! GetParameter( command->Parameters, "InitiatorName=", InitiatorName ) )
			{
				printf( "ERROR: InitiatorName MUST be present in Login Command\n" );
				LoginResponse( response, 0x0200 );
				return;
			}

			if ( ! GetParameter( command->Parameters, "SessionType=", Value ) )
			{
				printf( "ERROR: SessionType MUST be present in Login Command\n" );
				LoginResponse( response, 0x0200 );
				return;
			}
			SessionType = GetSessionType( Value );

			TargetName[0] = 0;

			if ( SessionType != Discovery )
			{
				if ( ! GetParameter( command->Parameters, "TargetName=", TargetName ) )
				{
					printf( "ERROR: TargetName MUST be present when SessionType is not Discovery\n" );
					LoginResponse( response, 0x0200 );
					return;
				}
			}

			if ( ParametersContainOperational( command->Parameters ) )
			{
				printf( "ERROR: MUST not have Operational Parameters in Login Command\n" );
				LoginResponse( response, 0x0200 );
				return;
			}

			// -----------------------------
			// Initial LOGIN checks Complete
			// -----------------------------

			if ( ParametersContainSecurity( command->Parameters ) ) 
			{
				// Login command had security parameters so Initiator wants to enter Security
				// The Reply MUST have Zero Or More security requests.
				TState = TSecurity;
			}
			else if ( TargetWantsSecurity )
			{
				// Login command had NO security parameters but target wants to enter Security Phase
				// Special case as the reply MUST have One Or More security requests.
				TState = TSecurity;
				ProcessInParameters( command->Parameters, TargetParamSet );

				printf( "\nRESPONSE:\n" );
				response->Type = TEXT;
				printf( "Enter One or More Security Requests\n" );
				ReadParameters( response->Parameters, TargetParamSet );
				if ( GetNoOfParameters( response->Parameters ) == 0 )
				{
					printf( "ERROR: MUST have at least one security parameter\n" );
					DoExit( 1 );
				}
				if ( ParametersContainOperational( response->Parameters ) )
				{
					printf( "ERROR: MUST NOT have Operational Parameters present is security phase\n" );
					DoExit( 1 );
				}
				if ( ParametersContainSecurity( response->Parameters ) == 0 )
				{
					printf( "ERROR: MUST have at least one secuirty parameter present\n" );
					DoExit( 1 );
				}
				response->FBit = SetFBit( &TargetFBit );
				return;
			}
			else
			{
				// NO security required by either side so goto operational state.
				TState = TOperational;
			}
			break;

		case TSecurity:
			ProcessInParameters( command->Parameters, TargetParamSet );

			if ( GetParameter( command->Parameters, SCC, NULL ) == 1 )
			{
				// SCC included in command so process it
				ChangeParameterState( SCC, TargetParamSet, STATUS );
				T_InitiatorSCC = 1;
			}
			else if ( T_InitiatorSCC == 1 )
			{
				printf( "ERROR: Initiator MUST report SecurityContextComplete=yes when Security Phase Complete\n" );
				DoExit ( 1 );
			}

			// --------------------
			// Create TEXT Response
			// --------------------
			response->Type = TEXT;
			printf( "\nRESPONSE:\n" );

			printf( "Enter %s Security Replies and %s Security Requests\n",
					( AreAllRepliesSatisfied( TargetParamSet ) ) ? "NO" : "all",
					( ( T_TargetSCC == 1 ) || ( TargetFBit == 1 ) ) ? "NO" : "zero or more"
				  );	

			ReadParameters( response->Parameters, TargetParamSet );

			if ( GetParameter( response->Parameters, SCC, NULL ) == 1 )
			{
				// Target is completing security phase 
				ChangeParameterState( SCC, TargetParamSet, STATUS );  // Change its type to status
				if ( AreAllRequestsSatisfied( TargetParamSet ) == 0 )
				{
					printf( "ERROR: Target MUST complete all target initiated negotiations before setting SecurityContextComplete=yes\n" );
					DoExit( 1 );
				}

				ChangeParameterState( SCC, TargetParamSet, STATUS );
				T_TargetSCC = 1;
			}
			else if ( T_TargetSCC == 1 )
			{
				printf( "ERROR: Target MUST report SecurityContextComplete=yes when Security Phase Complete\n" );
				DoExit ( 1 );
			}

			response->FBit = SetFBit( &TargetFBit );

			if ( ( T_InitiatorSCC == 1 ) &&
				 ( T_TargetSCC == 1 )    &&
				 ( GetNoOfParameters( response->Parameters ) == 1 )  // Only SCC in the response
			   )
			{
				// All Criteria for exiting secuirty phase have been met.
				printf ( "TARGET: Security Context Complete\n" );

				if ( ( command->FBit == 1 ) && ( response->FBit == 1 ) )
				{
					// Both sides do not wnat to nedotiate operational parameters
					TState = TLoggedIn;
					response->Type = LOGIN;
					response->LoginStatus = 0;
				}
				else
				{
					TState = TOperational;
				}
			}
			return;
			break;

		case TOperational:
			ProcessInParameters( command->Parameters, TargetParamSet );

			// --------------------
			// Create TEXT Response
			// --------------------
			response->Type = TEXT;
			printf( "\nRESPONSE:\n" );
			printf( "Enter %s Operational Replies and %s Operational Requests\n",
					( AreAllRepliesSatisfied( TargetParamSet ) ) ? "NO" : "all",
					( ( TargetFBit == 1 ) ) ? "NO" : "zero or more"
				  );	
			response->Parameters[0][0] = 0;
			if ( ( ! AreAllRepliesSatisfied( TargetParamSet ) ) ||  // Target has not satisfied all initiator requests
				 ( TargetFBit == 0 ) // Target is done!
			   )
			{
				// There are either initiator requests to reply to or more requests to make or both!
				ReadParameters( response->Parameters, TargetParamSet );
			}

			response->FBit = SetFBit( &TargetFBit );

			if ( ( response->FBit == 1 ) &&
				 ( command->FBit == 1 ) &&
				 ( AreAllRequestsSatisfied( TargetParamSet ) )
			   )
			{
				// Both target and initiator are done!
				response->Type = LOGIN;
				TState = TLoggedIn;
			}
			return;
			break;
		
		case TLoggedIn:
			return;
			break;

		default:
			printf( "ERROR : Invalid TState\n" );
			DoExit ( 1 );
			break;
		}
	}
}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to create a basic (no parameters) login response
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static void LoginResponse( PDU_t *response, int Status )
{
	response->Type = LOGIN;
	response->FBit = 1;
	TargetFBit = 1;
	response->LoginStatus = 0x200;
	response->Parameters[0][0] = 0;
}

// ++++++++++++++++++++++++++++++++++++++++++++
// Function to send an Operational Text Command
// ++++++++++++++++++++++++++++++++++++++++++++
static void SendOperationalTextCommand( PDU_t *command, PDU_t *response )
{
	int		AllRepliesSatisfied;

	command->Type = TEXT;
	printf( "\nCOMMAND:\n" );
	AllRepliesSatisfied = AreAllRepliesSatisfied( InitiatorParamSet );
	printf( "Enter %s Operational Responses and %s Operational Requests\n",
			( AllRepliesSatisfied == 1 ) ? "NO" : "all",
			( InitiatorFBit == 1 ) ? "NO" : "one or more"
		  );

	command->Parameters[0][0] = 0;
	if ( ( InitiatorFBit == 0 ) || ( AllRepliesSatisfied == 0 ) )
	{
		// Initiator is not done yest or it has to reply to target requests,
		ReadParameters( command->Parameters, InitiatorParamSet );
	}

	command->FBit = SetFBit( &InitiatorFBit );
	SendCommand( command, response );				
}


// +++++++++++++++++++++++++++++++++++++++++
// Function to send an Security Text Command
// +++++++++++++++++++++++++++++++++++++++++
static void SendSecurityTextCommand( PDU_t *command, PDU_t *response )
{
	command->Type = TEXT;
	printf( "\nCOMMAND:\n" );
	printf( "Enter %s Security Responses and %s Security Requests\n",
			( AreAllRepliesSatisfied( InitiatorParamSet ) ) ? "NO" : "all",
			( ( I_InitiatorSCC == 1 ) || ( InitiatorFBit == 1 ) ) ? "NO" : "zero or more"
		  );

	ReadParameters( command->Parameters, InitiatorParamSet );

	if ( GetParameter( command->Parameters, SCC, NULL ) == 1 )
	{
		// SCC is a parameter so process it
		ChangeParameterState( SCC, InitiatorParamSet, STATUS );  // Change its type to status
		if ( AreAllRequestsSatisfied( InitiatorParamSet ) == 0 )
		{
			printf( "ERROR: Initiator MUST complete all initiator initiated negotiations before setting SecurityContextComplete=yes\n" );
			DoExit( 1 );
		}
		I_InitiatorSCC = 1;
	}
	else if ( I_InitiatorSCC == 1 )
	{
		printf( "ERROR: Target MUST report SecurityContextComplete=yes when Security Phase Complete\n" );
		DoExit ( 1 );
	}

	command->FBit = SetFBit( &InitiatorFBit );

	SendCommand( command, response );
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to send a command, process it (by the target), and print the response
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static void SendCommand( PDU_t *command, PDU_t *response )
{
	int i;

	printf( "I->T : %s Command\n", PDUNames[command->Type] );
	printf( "       FBit=%d\n", command->FBit );

	i = 0;
	while ( command->Parameters[i][0] != 0 )
	{
		printf( "       %s\n", command->Parameters[i] );
		i ++;
	}

	TargetProcessCommand( command, response );
	PrintResponse( response );
}


// +++++++++++++++++++++++++++++++++++++++++++++
// Function to read and process the command line
// +++++++++++++++++++++++++++++++++++++++++++++
static void ReadString( const char *text, char *str )
{
	char	tempstr[MAX_PARAM_LENGTH];
	char	*ptr;
	char	*str2;

	do
	{
	    str2 = str;

	    printf( "Enter %s : ", text );
	    fgets( tempstr, MAX_PARAM_LENGTH, stdin );

	    if ( feof( stdin ) )
	    {
		// Handle EOF for input redirect
		printf( "\nERROR: End of file reached prematurely\n" );
		DoExit( 1 );
	    }

	    ptr = tempstr;
	    while ( ( *ptr != 0 ) && ( *ptr != '\n' ) && ( *ptr != '\r' ) )
	    {
		    // Strip out any spaces
		    if ( ! isspace( *ptr ) )
		    {
			*str2 = *ptr;
			str2 ++;
		    }

		    ptr ++;
	    }

	    *str2 = 0;


#if ECHOINPUT == 1
	    if ( str[0] == '#' )
	    {
		printf( "%s\n", str );
	    }
#endif
	
	} while ( str[0] == '#' ); // Ignore comments


#if ECHOINPUT == 1
	puts( str );
#endif
}

// ++++++++++++++++++++++++++++++++++++++++++
// Function to read and process a boolean
// Returns the boolean value (0 or 1 doe-doe!)
// ++++++++++++++++++++++++++++++++++++++++++
static int ReadBoolean( const char *text )
{
	char	str[MAX_PARAM_LENGTH];
	char	boolstr[MAX_PARAM_LENGTH];

	sprintf( str, "%s [0|1]", text );


	do
	{
		ReadString(  str, boolstr );

		if ( boolstr[0] == 0 )
		{
		    printf( "ERROR: Empty String for FBit\n" );
		    DoExit( 1 );
		}

	} while ( ( boolstr[0] != '0' ) && ( boolstr[0] != '1' ) );

	return (int) ( boolstr[0] - '0' );
}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to test if a parameter is a security parameter
// Returns 1 if it is and 0 otherwise
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++
static int IsSecurityParameter( char *parameter )
{
	int i;

	i = 0;

	while ( SecurityParameters[i] != NULL )
	{
		if ( memcmp( parameter, SecurityParameters[i], strlen( SecurityParameters[i] ) ) == 0 )
		{
			return 1;
		}

		i ++;
	}

	return 0;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to test if a command or response contains one or more security parameters
// Returns 1 if it is and 0 otherwise
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static int ParametersContainSecurity( PARAMS parameters )
{
	int		i;

	i = 0;
	while ( parameters[i][0] != 0 )
	{
		if ( IsSecurityParameter( parameters[i] ) )
		{
			return 1;
		}
		i ++;
	}
	
	return 0;
}


// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to test if a parameter is an operational parameter
// Returns 1 if it is and 0 otherwise
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static int IsOperationalParameter( char *parameter )
{
	int i;

	i = 0;

	while ( OperationalParameters[i] != NULL )
	{
		if ( memcmp( parameter, OperationalParameters[i], strlen( OperationalParameters[i] ) ) == 0 )
		{
			return 1;
		}

		i ++;
	}

	return 0;
}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to test if a command or response contains one or more operational parameters
// Returns 1 if it is and 0 otherwise
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static int ParametersContainOperational( PARAMS parameters )
{
	int		i;

	i = 0;
	while ( parameters[i][0] != 0 )
	{
		if ( IsOperationalParameter( parameters[i] ) )
		{
			return 1;
		}
		i ++;
	}
		
	return 0;
}


// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to test if a parameter exists and returns the string value of the parameter
// Returns 1 if it exists and 0 otherwise
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static int GetParameter( PARAMS parameters, char *parameter, char *value )
{
	int		i;
	char	*eq;

	// Clean Up
	i = 0;
	while ( parameters[i][0] != 0 )
	{
		if ( memcmp( parameters[i], parameter, strlen( parameter ) ) == 0 )
		{
			if ( value != NULL )
			{
				eq = strchr( parameters[i], '=' );
				if ( eq != NULL )
				{
					eq ++;  // Move past equals
					strcpy( value, eq );
				}
				else
				{
					printf( "ERROR: Parameter Expected : %s\n", parameters[i] );
					DoExit( 0 );
				}
			}
			return 1;
		}
	
		i ++;
	}
		
	return 0;
}


// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to get the number of parameters in a command or response
// Returns the number of parameters
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static int GetNoOfParameters( PARAMS parameters )
{
	int		i;

	i = 0;
	while ( parameters[i][0] != 0 )
	{	
		i ++;
	}
		
	return i;
}

// ++++++++++++++++++++++++++++
// Function to print a response
// ++++++++++++++++++++++++++++
static void PrintResponse( PDU_t *response )
{
	int	i;

	printf( "T->I : %s Response\n", PDUNames[response->Type] );
	printf( "       FBit=%d\n", response->FBit );

	if ( response->Type == LOGIN )
	{
		printf( "       Status=%04X\n", response->LoginStatus );
	}

	i = 0;
	while ( response->Parameters[i][0] != 0 )
	{
		printf( "       %s\n", response->Parameters[i] );
		i ++;
	}
}


// ++++++++++++++++++++++++++++++++++++++++++++
// Function to read a number of text parameters
// ++++++++++++++++++++++++++++++++++++++++++++
static void ReadParameters( PARAMS Parameters, PARAMTYPE *ParamSet )
{
	int		i;

	for ( i = 0; i < MAX_PARAMETERS; i ++ )
	{
		ReadString( "Text Parameter [Return to End]", Parameters[i] );
		if ( Parameters[i][0] == 0 )
		{
			return;
		}

		AddOutParameter( Parameters[i], ParamSet );
	}

}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to convert a text string into the sessiontype
// Returns the session type
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++
static SESSIONTYPE GetSessionType( char *value )
{
	if ( strcmp( value, "Discovery" ) == 0 )
	{
		return Discovery;
	}
	
	if ( strcmp( value, "Normal" ) == 0 )
	{
		return Normal;
	}

	if ( strcmp( value, "CopyManager" ) == 0 )
	{
		return CopyManager;
	}

	if ( strcmp( value, "Boot" ) == 0 )
	{
		return Boot;
	}

	printf( "ERROR : Invalid Session Type\n" );
	DoExit( 1 );

	return Normal;
}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to add an outgoing parameter to the associated parameter set
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static void AddOutParameter( char *Parameter, PARAMTYPE *ParamSet )
{
	int		i;
	char	*tempParam;
	char	*ptr;

	tempParam = strdup( Parameter );
	ptr = strchr( tempParam, '=' );
	if ( ptr != NULL )
	{
		ptr ++; // Move past =
		*ptr = 0;
	}
	
	i = 0;
	while ( ParamSet[i].State != FREE )
	{
		if ( strcmp( tempParam, ParamSet[i].Text ) == 0 )
		{
			// Eureka!
			if ( ParamSet[i].State == OUT_REQUEST )
			{
				// Already a request
				printf ( "Warning: Parameter %s already requested\n", tempParam );
				return;
			}

			if ( ParamSet[i].State == IN_REQUEST )
			{
				// Request from initiator 
				ParamSet[i].State = DONE;
				return;
			}

			if ( ParamSet[i].State == DONE )
			{
				// Already a request
				printf ( "Warning: Parameter %s already negotiated\n", tempParam );
				return;
			}

			if ( ParamSet[i].State == STATUS )
			{
				return;
			}
		}

		i ++;
	}

	// Param Not Found: Add it - it MUST be a request!
	ParamSet[i].State = OUT_REQUEST;
	strcpy( ParamSet[i].Text, tempParam );
}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to add an incoming parameter to the associated parameter set
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static void AddInParameter( char *Parameter, PARAMTYPE *ParamSet )
{
	int		i;
	char	*tempParam;
	char	*ptr;

	tempParam = strdup( Parameter );
	ptr = strchr( tempParam, '=' );
	if ( ptr != NULL )
	{
		ptr ++; // Move past =
		*ptr = 0;
	}
	
	i = 0;
	while ( ParamSet[i].State != FREE )
	{
		if ( strcmp( tempParam, ParamSet[i].Text ) == 0 )
		{
			// Eureka!
			if ( ParamSet[i].State == OUT_REQUEST )
			{
				// a request needing a reply: this is it
				ParamSet[i].State = DONE;
				return;
			}

			if ( ParamSet[i].State == IN_REQUEST )
			{
				// Request from initiator 
				printf ( "Warning: Parameter %s already requested\n", tempParam );
				return;
			}

			if ( ParamSet[i].State == DONE )
			{
				// Already a request
				printf ( "Warning: Parameter %s already negotiated\n", tempParam );
				return;
			}

			if ( ParamSet[i].State == STATUS )
			{
				return;
			}
		}

		i ++;
	}

	// Param Not Found: Add it - it MUST be a request!
	ParamSet[i].State = IN_REQUEST;
	strcpy( ParamSet[i].Text, tempParam );
}


// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to process the parameters of a received command or response
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static void ProcessInParameters( PARAMS Parameters, PARAMTYPE *ParamSet )
{
	int i;

	i = 0;
	while ( Parameters[i][0] != 0 )
	{
		AddInParameter( Parameters[i], ParamSet );
		i ++;
	}
	ChangeParameterState( "InitiatorName=", ParamSet, DONE );
	ChangeParameterState( "SessionType=", ParamSet, DONE );
	ChangeParameterState( "TargetName=", ParamSet, DONE );

	if ( ! CheckAllRequestsSatisfied( ParamSet ) )
	{
		DoExit( 1 );
	}
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to check that all requests have been replied to
// Returns 1 if all replied to and 0 otherwise
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static int CheckAllRequestsSatisfied( PARAMTYPE *ParamSet )
{
	int		i;

	i = 0;
	while ( ParamSet[i].State != FREE )
	{
		if ( ParamSet[i].State == OUT_REQUEST )
		{
			printf( "ERROR: Unsatisfied parameter %s\n", ParamSet[i].Text );
			return 0;  // Unsatisfied Parameter
		}

		i ++;
	}

	return 1; // All Okay
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to test that all requests have been replied to
// As CheckAllRequestsSatisfied but no print out
// Returns 1 if all replied to and 0 otherwise
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static int AreAllRequestsSatisfied( PARAMTYPE *ParamSet )
{
	int		i;

	i = 0;
	while ( ParamSet[i].State != FREE )
	{
		if ( ParamSet[i].State == OUT_REQUEST )
		{
			return 0;  // Unsatisfied Parameter
		}

		i ++;
	}

	return 1; // All Okay
}


// ++++++++++++++++++++++++++++++++++++++++++++
// Function to test for any outstanding replies
// Returns 1 if all replied to and 0 otherwise
// ++++++++++++++++++++++++++++++++++++++++++++
static int AreAllRepliesSatisfied( PARAMTYPE *ParamSet )
{
	int		i;

	i = 0;
	while ( ParamSet[i].State != FREE )
	{
		if ( ParamSet[i].State == IN_REQUEST )
		{
			return 0;  // Unsatisfied Parameter
		}

		i ++;
	}

	return 1; // All Okay
}


// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function to change the state of a item in the parameter set
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static void ChangeParameterState( char *Parameter, PARAMTYPE *ParamSet, PSTATE state )
{
	int		i;

	i = 0;
	while ( ParamSet[i].State != FREE )
	{
		if ( strcmp( Parameter, ParamSet[i].Text ) == 0 )
		{
			// Eureka!
			ParamSet[i].State = state;
			return;
		}
		i ++;
	}
}

// ++++++++++++++++++++++++++++++++++++++++++++++++
// Function to exit: only need a single break point
// ++++++++++++++++++++++++++++++++++++++++++++++++
static void DoExit( int code )
{
	exit( code );
}


// +++++++++++++++++++++++++++++++++++++++++
// Function to set the FBit (if it is not 1)
// The current value of the FBit
// +++++++++++++++++++++++++++++++++++++++++
static int SetFBit( int *FBit )
{
	if ( *FBit == 0 )
	{
		*FBit = ReadBoolean( "FBit" );
	}
	return *FBit;
}

